
from bridger.endpoints.metadata_config import EndpointConfig
from rest_framework.reverse import reverse

class ClubHouseEndpointConfig(EndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(f"bridger:clubhouse-list", request=self.request)

    def get_instance_endpoint(self, **kwargs):
        return self.get_list_endpoint()

    def get_create_endpoint(self, **kwargs):
        return self.get_list_endpoint()


from typing import Dict

import requests
from django.utils.html import strip_tags
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse

from bridger.enums import Button
from typing import Optional
from bridger import serializers
from bridger.settings import bridger_settings
from bridger.viewsets import ViewSet, ModelViewSet

from .display import ClubHouseDisplayConfig
from .endpoints import ClubHouseEndpointConfig
from .title import ClubHouseTitleConfig

story_type_choices = (("bug", "Bug"), ("feature", "Feature"))


def config(request: Request) -> Dict:
    return {
        "endpoint": reverse("bridger:clubhouse-list", request=request),
        "add": reverse("bridger:clubhouse-list", request=request),
    }


class ClubhouseSerializer(serializers.Serializer):
    id = serializers.PrimaryKeyField()
    name = serializers.CharField(label="Name")
    story_type = serializers.ChoiceField(label="Type", choices=story_type_choices)
    description = serializers.TextField(required=False)
    created_at = serializers.DateTimeField(read_only=True)
    completed = serializers.BooleanField(read_only=True)

    def validate(self, data):
        data["name"] = strip_tags(data["name"])
        description = f"**Reported by: {self.context['request'].user.username}**\n\n{data['description']}"
        data["description"] = description

        return data

    def create(self, validated_data):
        assert bridger_settings.CLUBHOUSE_API_TOKEN
        assert bridger_settings.CLUBHOUSE_PROJECT_ID

        response = requests.post(
            f"https://api.clubhouse.io/api/v3/stories?token={bridger_settings.CLUBHOUSE_API_TOKEN}",
            json={**validated_data, "project_id": bridger_settings.CLUBHOUSE_PROJECT_ID,},
            headers={"Content-Type": "application/json"},
        )

        return self.__class__(response.json())

   
class ClubHouseView(ViewSet):
    IDENTIFIER = "clubhouse"
    title_config_class = ClubHouseTitleConfig
    endpoint_config_class = ClubHouseEndpointConfig
    display_config_class = ClubHouseDisplayConfig

    BUTTONS = [Button.REFRESH.value, Button.NEW.value]
    CREATE_BUTTONS = [Button.SAVE.value]
    
    serializer_class = ClubhouseSerializer

    # We need to add this otherwise DjangoModelPermissions complains it cannot find a get_queryset method
    permission_classes = (AllowAny,)

    def create(self, request):
        serializer = self.serializer_class(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)
        clubhouse_story = serializer.save()
        return Response({"instance": clubhouse_story.data}, status=201)

    def retrieve(self, request, pk=None):
        assert bridger_settings.CLUBHOUSE_API_TOKEN

        response = requests.get(
            f"https://api.clubhouse.io/api/v3/stories/{pk}?token={bridger_settings.CLUBHOUSE_API_TOKEN}",
            headers={"Content-Type": "application/json"},
        )

        if response.status_code == 404:
            return Response({}, status=404)
        json_response = response.json()
        serializer = self.serializer_class(json_response)
        return Response({"instance": serializer.data})

    def list(self, request):
        assert bridger_settings.CLUBHOUSE_API_TOKEN
        assert bridger_settings.CLUBHOUSE_PROJECT_ID

        response = requests.get(
            f"https://api.clubhouse.io/api/v3/projects/{bridger_settings.CLUBHOUSE_PROJECT_ID}/stories?token={bridger_settings.CLUBHOUSE_API_TOKEN}",
            headers={"Content-Type": "application/json"},
        )
        json_response = response.json()

        serializer = self.serializer_class(json_response, many=True)
        return Response({"results": serializer.data})



from channels.db import database_sync_to_async
from django.contrib.auth import get_user_model
from django.utils import timezone

from bridger.websockets.consumers import AsyncAuthenticatedJsonWebsocketConsumer

from .models import Notification


@database_sync_to_async
def get_notification(notification_id):
    return Notification.objects.get(id=notification_id)


@database_sync_to_async
def update_notification(notification_id):
    Notification.objects.filter(id=notification_id).update(timestamp_received=timezone.now())


class NotificationConsumer(AsyncAuthenticatedJsonWebsocketConsumer):
    async def connect(self):
        try:
            if not self.scope.get('user', None):
                raise ValueError("Scope does not have user attribute")
            await self.accept()
            channel_layer_name = f"notification-{self.scope['user'].id}"
            await self.channel_layer.group_add(channel_layer_name, self.channel_name)
        except ValueError as e:
            await self.close(code=403)

    async def get_notification_message(self, user):
        num_unreceived, num_unread = await database_sync_to_async(self.get_notification_information)(user)
        return f"You have {num_unreceived} new notifications and {num_unread} unread notifications."

    def get_notification_information(self, user):
        qs = Notification.objects.filter(recipient=user)
        num_unreceived = qs.filter(timestamp_received__isnull=True).update(timestamp_received=timezone.now())
        num_unread = qs.filter(timestamp_read__isnull=True).count()

        return num_unreceived, num_unread

    async def notification_info(self, content):
        notification_message = await self.get_notification_message(get_user_model().objects.get(id=content["user_id"]))
        await self.send_json({"title": "Notification", "message": notification_message})

    async def notification_notify(self, content):
        notification = await get_notification(notification_id=content["notification_id"])
        await self.send_json(notification.to_payload())

    async def receive_json(self, content):
        """ Receives a dict in format of: {"notification_id": <int>} and marks it as received
        """
        await update_notification(notification_id=content["notification_id"])



from channels.auth import UserLazyObject
from channels.db import database_sync_to_async
from channels.middleware import BaseMiddleware
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import close_old_connections
from jwt import decode as jwt_decode
from rest_framework_simplejwt.exceptions import InvalidToken, TokenError
from rest_framework_simplejwt.tokens import UntypedToken


@database_sync_to_async
def get_user(user_id):
    try:
        return get_user_model().objects.get(id=user_id)
    except User.DoesNotExist:
        return AnonymousUser()


class JWTAuthMiddleware(BaseMiddleware):
    """ Auth Middleware that authenticates a user with a JWT TOKEN """

    # def __init__(self, inner):
    #     self.inner = inner

    def populate_scope(self, scope):
        if "user" not in scope:
            scope["user"] = UserLazyObject()

    async def resolve_scope(self, scope):

        # We should close all old connections here, somehow this bites itself with pytest
        # close_old_connections()

        try:
            jwt_access_token = scope["cookies"]["JWT-access"]
            UntypedToken(jwt_access_token)
        except (InvalidToken, TokenError, KeyError):
            return self.inner(dict(scope))
        else:
            decoded_data = jwt_decode(jwt_access_token, settings.SECRET_KEY, algorithms=["HS256"])
            scope["user"]._wrapped = await get_user(user_id=decoded_data["user_id"])

        # return self.inner(dict(scope, user=user))



from bridger.titles.metadata_config import TitleConfig

class ClubHouseTitleConfig(TitleConfig):
    def get_instance_title(self):
        return "Bug Report"
    def get_list_title(self):
        return "Bug Reports"
    def get_create_title(self):
        return "New Report"
    def get_delete_title(self):
        return "Delete Report"
